#include <iostream>
#include <fstream>
#include <vector>
#include <cassert>
	
template<typename T>
T dp(const std::vector<std::pair<T,T>> &A,
           std::vector<std::vector<std::pair<T,T>>> &C) {
  int p=A.size(); 
  for (int j=0;j<p;j++)
    for (int i=j;i>=0;i--) 
      if (i==j) 
        C[j][i]=std::make_pair(0,0);
      else 
        for (int k=i+1;k<=j;k++) {
          T cost=C[j][k].first+C[k-1][i].first+A[j].first*A[k].second*A[i].second;
          if (k==i+1||cost<C[j][i].first) C[j][i]=std::make_pair(cost,k);;
        }
  return C[p-1][0].first;
}

int main(int argc, char* argv[]) {
  assert(argc==2); std::ifstream in(argv[1]);
  using T=unsigned long;
  // Matrix Chain Product as sequence of m x n factors 
  int p; in >> p; assert(p>0); 
  std::vector<std::pair<T,T>> A(p,std::make_pair(0,0)); 
  for (int i=0;i<p;i++) in >> A[i].first >> A[i].second; 
  std::cout << "Matrix Chain Product:" << std::endl;
  for (int i=p-1;i>=0;i--) 
    std::cout << "A[" << i << "]: " << A[i].first << " x " << A[i].second << std::endl;
  // Dynamic Programming Table as // p x p lower triangular matrix 
  // storing optimal cost and split position per subchain
  std::vector<std::vector<std::pair<T,T>>> 
    C(p,std::vector<std::pair<T,T>>(p,std::make_pair(0,0))); 
  dp(A,C);
  std::cout << "Dynamic Programming Table:" << std::endl;
  for (int j=0;j<p;j++) 
    for (int i=j-1;i>=0;i--) {
      std::cout << "Cost(C[" << j << "][" << i << "])=" 
           << C[j][i].first << "; " 
           << "Split before A[" << C[j][i].second << "]; " << std::endl;
    }
  return 0;
}
